from pantea.potentials.nnp.potential import NNP, NeuralNetworkPotential
from pantea.potentials.nnp.settings import NeuralNetworkPotentialSettings

__all__ = [
    "NeuralNetworkPotential",
    "NeuralNetworkPotentialSettings",
    "NNP",
]
